const http = require('http');
const request = require('request-promise');
const { Agent: HttpAgent } = require('http');
const { Agent: HttpsAgent } = require('https');
const is_ip_private = require('private-ip');
const errorPrefix = 'SSRF:';
const patchAgent = ({
                        isPrivate = (address) => is_ip_private(address),
                        agent,
                    },
) => {
    const createConnection = agent.createConnection;
    agent.createConnection = function (options, fn) {
        const { host: address } = options;
        if (isPrivate(address)) {
            throw new Error(`${errorPrefix} private address ${address} is not allowed.`);
        }

        const client = createConnection.call(this, options, fn);
        client.on('lookup', (err, address) => {
            if (err || !isPrivate(address)) {
                return;
            }

            return client.destroy(new Error(`${errorPrefix} DNS lookup of private '${client._host}' returned ${address} is not allowed.`));
        });

        return client;
    };
    agent.PATCHED = true;
    return agent;
};


const httpAgent = patchAgent({ agent: new HttpAgent() });
const httpsAgent = patchAgent({ agent: new HttpsAgent() });

/**
 * Prepare Ssrf filter agent to use in request options as as agent.
 * The agent can be used directly by any kind of client
 * @param url
 * @param ssrf
 * @param allowListDomains
 * @param trace
 * @returns {undefined|*}
 */
const getAgent = ({ url, ssrf = true, allowListDomains = [], trace = false }) => {
    if (!ssrf) {
        return undefined;
    }
    const urlObject = new URL(url);
    const protocol = urlObject.protocol;
    const hostname = urlObject.hostname;
    if (allowListDomains.includes(hostname)) {
        trace && console.log(`Allow list match: ${hostname}, in: ${allowListDomains}, ignore ssrf`);
        return undefined;
    }
    if (protocol === 'https:') return httpsAgent;
    if (protocol === 'http:') return httpAgent;
    new Error(`${errorPrefix} Bad protocol, url must start with http/https, Got ${url}`);
};

/**
 * method returning options with agent based on url parameter
 */
function ssrfOptions({ url, options = {}, trace = true, allowListDomains = [] }) {
    const agent = getAgent({ url, allowListDomains });
    return { ...options, agent };
}

/**
 * convenience method for creating request options.
 * returning request options with agent for ssrf safe requests.
 * for enriching options with ssrf agent filter,
 * '.uri' is required in options or provided as 'url' explicitly
 */
function requestSsrfOptions({
                                url = undefined,
                                options = { uri: undefined },
                                trace = false,  // for test/debug
                                ssrf = true,    // for test/debug
                                allowListDomains = [], //
                            }) {
    url = url || options.uri;
    options = ssrf ? ssrfOptions({ url: url, options, trace, allowListDomains }) : options;
    return { ...options, uri: url };
}

/**
 * Native node http lib use of ssrf agent
 */
async function httpSsrfGet({ url, trace = true, ssrf = false, allowListDomains = [] }) {
    trace && console.log(`Calling ${url} ssrf:${ssrf} allowListDomains:${allowListDomains}`);
    const options = {
        agent: getAgent({ url, ssrf, allowListDomains, trace }),
    };
    const waitfor = new Promise((resolve, reject) => {
        let data = [];
        trace && console.log(`http.get ${url},  ${JSON.stringify(options)}`);
        http.get(url, options, res => {
            const headerDate = res.headers && res.headers.date ? res.headers.date : 'no response date';
            trace && console.log('Status Code:', res.statusCode);
            trace && console.log('Date in Response header:', headerDate);

            res.on('data', chunk => {
                data.push(chunk);
            });

            res.on('end', () => {
                trace && console.log('Response ended: ');
                resolve({ data: data.join(''), statusCode: res.statusCode });
            });
        })
            .on('error', err => {
                trace && console.log('Error: ', err.message);
                reject(err);
            });
    });
    const result = await waitfor;
    trace && console.log(`Called ${url} return ${JSON.stringify(result)}`);
    return result;
}

/**
 * convenience method for doing request get.
 * @param url
 * @param trace
 * @param ssrf
 * @param allowListDomains
 * @returns {Promise<*>}
 */
async function requestSsrfGet({ url, trace = true, ssrf = true, allowListDomains = [] }) {
    trace && console.log(`requestGet Calling ${url} ssrf: ${ssrf} allowListDomains:${allowListDomains}`);
    const options = requestSsrfOptions({ url, trace, ssrf, allowListDomains });
    try {
        const result = await request(options);
        trace && console.log(`requestGet Called ${url} return ${JSON.stringify(result)}`);
        return result;
    } catch (err) {
        trace && console.log('Error: ', err.message);
        throw err;
    }
}

/**
 * method for checking if an error was thrown by the Ssrf agent filter.
 * @param err
 * @returns {boolean}
 */
function isSsrfError(err) {
    return err.message.startsWith(errorPrefix);
}

module.exports = { getAgent, isSsrfError, requestSsrfOptions, requestSsrfGet, httpSsrfGet };
